# Native imports
from base64 import b64decode
from decimal import Decimal


def dynamo_object_to_dict(obj):
    """
    Converts a DynamoDB object (dictionary-like) to a native Python dictionary.

    @param obj: The object to convert.
    @return: A native Python dictionary representing the data.
    """

    def _simple_types(k, v):
        # Handle the string case
        if k == "S":
            value = v
        # Handle the numeric case
        elif k == "N":
            if "." in v:
                value = float(v)
            else:
                value = int(v)
        # Handle the binary case
        elif k == "B":
            value = b64decode(v)
        # Handle the string set case
        elif k == "SS":
            value = [s for s in v]
        # Handle the number set case
        elif k == "NS":
            value = [float(s) for s in v]
        # Handle the binary set case
        elif k == "BS":
            value = [b64decode(s) for s in v]
        # Handle the null case
        elif k == "NULL":
            value = None
        # Handle the boolean case
        elif k == "BOOL":
            value = v
        else:
            return False, None
        return True, value

    def _raise(dynamo_type):
        raise Exception("Unsupported DynamoDB type of {}".format(dynamo_type))

    items = {}
    for root_key, root_value in obj.items():
        # Handle bottom-most level
        if not isinstance(root_value, dict):
            parsed, parsed_value = _simple_types(root_key, root_value)
            if parsed:
                return parsed_value
            _raise(root_key)
        # Handle lists of dicts
        elif root_key == "M":
            return dynamo_object_to_dict(root_value)
        # Handle nested dicts
        else:
            for k, v in root_value.items():
                # Handle the simple cases
                parsed, value = _simple_types(k, v)
                if not parsed:
                    # Handle the map/dictionary case
                    if k == "M":
                        value = dynamo_object_to_dict(v)
                    # Handle the list case
                    elif k == "L":
                        value = [dynamo_object_to_dict(o) for o in v]
                    # Handle the unsupported case
                    else:
                        _raise(k)
                items[root_key] = value
    return items


def dynamo_dict_to_dict(obj):
    """
    Convert a dynamo compatible dictionary to an equivalent python dictionary.

    CAUTION: This method is very basic and presently only handles floats. Add to this as needed.
     Note that it is best to encode to JSON then use this to perform cleanup.

    @param obj: A dynamo-compatible dictionary.
    @return: A dict.
    """
    if isinstance(obj, dict):
        return {k: dynamo_dict_to_dict(v) for k, v in obj.items()}
    elif isinstance(obj, Decimal):
        if "." in str(obj):
            return float(obj)
        return int(obj)
    elif isinstance(obj, list):
        return [dynamo_dict_to_dict(list_item) for list_item in obj]
    return obj


def dict_to_dynamo_dict(obj):
    """
    Convert a Python dictionary to an equivalent dynamo compatible dictionary.

    CAUTION: This method is very basic and presently only handles floats. Add to this as needed. Note that it is best to
    encode to JSON then use this to perform cleanup.

    @param obj: A dictionary.
    @return: A dynamo compatible dict.
    """

    if isinstance(obj, dict):
        return {k: dict_to_dynamo_dict(v) for k, v in obj.items()}
    elif isinstance(obj, float):
        return Decimal(str(obj))
    elif isinstance(obj, list):
        return [dict_to_dynamo_dict(list_item) for list_item in obj]
    return obj
