import base64
import json
import struct
from .s3_client import S3Client

try:
    from Crypto import Random
    from Crypto.Cipher import AES
    from io import BytesIO
except Exception:
    pass


class KmsClient(object):
    def __init__(
        self,
        boto_session,
        boto_config=None,
        kms_endpoint_url=None,
        s3_endpoint_url=None,
    ):
        """

        @param boto_session: Pre-existing boto3 session
        @param boto_config: Optional boto config for the kms client
        """

        self.kms_client = boto_session.client(
            "kms", config=boto_config, endpoint_url=kms_endpoint_url
        )
        self.s3_instance = S3Client(
            boto_session=boto_session,
            boto_config=boto_config,
            endpoint_url=s3_endpoint_url,
        )

    def decrypt(self, cipher_text, encrypt_context=None):
        """
        Decrypts the given ciphertext

        @param cipher_text: Value to be decrypted
        @param encrypt_context: Optional name-value pair that specifies the encryption context
        @return: The decrypted cipher text
        """

        return self.kms_client.decrypt(
            CiphertextBlob=cipher_text, EncryptionContext=encrypt_context
        )

    def decrypt_bytes(self, cipher_text):
        """
        Decrypt the given string

        @param cipher_text: Value to be decrypted
        @return: Decrypted string value
        """

        decoded_cipher_text = base64.b64decode(cipher_text)
        response = self.kms_client.decrypt(CiphertextBlob=decoded_cipher_text)
        return response["Plaintext"]

    def decrypt_string(self, cipher_text):
        """
        Decrypt the given string

        @param cipher_text: Value to be decrypted
        @return: Decrypted string value
        """

        return self.decrypt_bytes(cipher_text).decode("utf-8")

    def encrypt(self, key_alias, plaintext):
        """
        Encrypts plaintext into ciphertext by using a customer master key

        @param key_alias: A unique identifier for the customer master key
        @param plaintext: Data to be encrypted
        @return: Encrypted plaintext
        """

        return self.kms_client.encrypt(KeyId=key_alias, Plaintext=plaintext)

    def encrypt_string(self, plaintext, key_alias):
        """
        Returns base64 encoded encrypted string

        @param plaintext: Text to be encrypted
        @param key_alias: A unique identifier for the customer master key
        @return: encrypted base 64 encoded string
        """
        response = self.kms_client.encrypt(KeyId=key_alias, Plaintext=plaintext)
        return base64.b64encode(response["CiphertextBlob"])

    def generate_data_key(self, key_alias, encrypt_context=None, key_spec=None):
        """
        Returns a data encryption key that can be used to encrypt data locally

        @param key_alias: The identifier of the CMK under which to generate and encrypt the data encryption key
        @param encrypt_context: A set of key-value pairs that represents additional authenticated data
        @param key_spec: A string describing the length of the data encryption key. 'AES_128' or 'AES_256'
        @return: data encryption key that can be used to encrypt data locally
        """
        return self.kms_client.generate_data_key(
            KeyId=key_alias, EncryptionContext=encrypt_context, KeySpec=key_spec
        )

    def encrypt_file_with_boto(self, file_name, s3_bucket_name, s3_key_name, key_alias):
        with open(file_name, "rb") as binary_file:
            file_contents = binary_file.read()

        self.encrypt_data_with_boto(
            file_contents, s3_bucket_name, s3_key_name, key_alias
        )

    def encrypt_data_with_boto(
        self, plaintext_str, s3_bucket_name, s3_key_name, key_alias
    ):
        unencrypted_file_size = len(plaintext_str)
        encrypt_ctx = {"kms_cmk_id": key_alias}
        key_data = self.generate_data_key(key_alias, encrypt_ctx, "AES_256")
        new_iv = Random.new().read(AES.block_size)

        encrypted_bytes = self.encrypt_file_in_memory(
            key_data["Plaintext"], new_iv, plaintext_str
        )
        encrypted_file = BytesIO(encrypted_bytes)

        self.s3_instance.store_encrypted_file_with_boto(
            key_data["CiphertextBlob"],
            new_iv,
            encrypt_ctx,
            encrypted_file,
            unencrypted_file_size,
            s3_bucket_name,
            s3_key_name,
        )

    def decrypt_file_with_boto(self, file_name, s3_bucket_name, s3_key_name):
        decrypted_bytes = self.decrypt_data_with_boto(s3_bucket_name, s3_key_name)

        with open(file_name, "wb") as binary_file:
            binary_file.write(decrypted_bytes)

    def decrypt_data_with_boto(self, s3_bucket_name, s3_key_name):
        # TODO: does this belong in kmsclient? Or is this response specific?
        #  Also should KMS rely on s3 or the other way around? Reevaluate in later ticket with scope.
        object_info = self.s3_instance.s3_client.head_object(
            Bucket=s3_bucket_name, Key=s3_key_name
        )
        metadata = object_info["Metadata"]

        envelope_key = base64.b64decode(metadata["x-amz-key-v2"])
        envelope_iv = base64.b64decode(metadata["x-amz-iv"])
        encrypt_ctx = json.loads(metadata["x-amz-matdesc"])
        original_size = metadata["x-amz-unencrypted-content-length"]

        decrypted_envelope_key = self.decrypt(
            cipher_text=envelope_key, encrypt_context=encrypt_ctx
        )

        data = BytesIO()
        self.s3_instance.download_fileobj(
            file_obj=data, bucket_name=s3_bucket_name, key_name=s3_key_name
        )
        decrypted_bytes = self.decrypt_file_in_memory(
            decrypted_envelope_key["Plaintext"],
            envelope_iv,
            int(original_size),
            data.getvalue(),
        )

        return decrypted_bytes

    @staticmethod
    def encrypt_file_in_memory(key, iv, plaintext_str, chunksize=16 * 1024):
        plaintext_io = BytesIO()
        plaintext_io.write(plaintext_str)
        plaintext_io.seek(0)
        encryptor = AES.new(key, AES.MODE_CBC, iv)
        ciphertext_io = BytesIO()

        while True:
            chunk = plaintext_io.read(chunksize)
            if len(chunk) == 0:
                break
            elif len(chunk) % 16 != 0:
                length = 16 - (len(chunk) % 16)
                chunk += struct.pack("B", length) * length
            ciphertext_io.write(encryptor.encrypt(chunk))

        return ciphertext_io.getvalue()

    @staticmethod
    def decrypt_file_in_memory(
        key, iv, original_size, ciphertext_str, chunksize=16 * 1024
    ):
        ciphertext_io = BytesIO()
        ciphertext_io.write(ciphertext_str)
        ciphertext_io.seek(0)
        decryptor = AES.new(key, AES.MODE_CBC, iv)
        plaintext_io = BytesIO()

        while True:
            chunk = ciphertext_io.read(chunksize)
            if len(chunk) == 0:
                break
            plaintext_io.write(decryptor.decrypt(chunk))
        plaintext_io.truncate(original_size)

        return plaintext_io.getvalue()
