from sqlalchemy import (
    BigInteger,
    CHAR,
    Column,
    DateTime,
    ForeignKey,
    Integer,
    SmallInteger,
    String,
    text,
)
from sqlalchemy.orm import relationship
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()
metadata = Base.metadata


class Operation(Base):
    __tablename__ = "operation"

    id = Column(SmallInteger, primary_key=True)
    name = Column(String(40))
    date_created = Column(DateTime(True), nullable=False, server_default=text("now()"))
    date_modified = Column(DateTime(True), nullable=False, server_default=text("now()"))


class Package(Base):
    __tablename__ = "package"

    id = Column(
        BigInteger,
        primary_key=True,
        server_default=text("nextval('package_id_seq'::regclass)"),
    )
    name = Column(String(256), nullable=False)
    sha256 = Column(CHAR(64), nullable=False)
    sha1 = Column(CHAR(40))
    md5 = Column(CHAR(32))
    date_created = Column(DateTime(True), nullable=False, server_default=text("now()"))
    date_modified = Column(DateTime(True), nullable=False, server_default=text("now()"))


class Source(Base):
    __tablename__ = "source"

    id = Column(
        Integer,
        primary_key=True,
        server_default=text("nextval('source_id_seq'::regclass)"),
    )
    name = Column(String(128), nullable=False)
    description = Column(String(256))
    date_created = Column(DateTime(True), nullable=False, server_default=text("now()"))
    date_modified = Column(DateTime(True), nullable=False, server_default=text("now()"))


class Status(Base):
    __tablename__ = "status"

    id = Column(SmallInteger, primary_key=True)
    name = Column(String(40))
    date_created = Column(DateTime(True), nullable=False, server_default=text("now()"))
    date_modified = Column(DateTime(True), nullable=False, server_default=text("now()"))


class DropboxItem(Base):
    __tablename__ = "dropbox_item"

    id = Column(
        BigInteger,
        primary_key=True,
        server_default=text("nextval('dropbox_item_id_seq'::regclass)"),
    )
    source_id = Column(ForeignKey("source.id"))
    name = Column(String(128), nullable=False)
    sha256 = Column(CHAR(64), nullable=False)
    date_created = Column(DateTime(True), nullable=False, server_default=text("now()"))
    date_modified = Column(DateTime(True), nullable=False, server_default=text("now()"))

    source = relationship("Source")


class PackageOperation(Base):
    __tablename__ = "package_operation"

    id = Column(
        BigInteger,
        primary_key=True,
        server_default=text("nextval('package_operation_id_seq'::regclass)"),
    )
    package_id = Column(ForeignKey("package.id"))
    operation_id = Column(ForeignKey("operation.id"))
    status_id = Column(ForeignKey("status.id"))
    date_created = Column(DateTime(True), nullable=False, server_default=text("now()"))
    date_modified = Column(DateTime(True), nullable=False, server_default=text("now()"))
    operation_date = Column(
        DateTime(True), nullable=False, server_default=text("now()")
    )

    operation = relationship("Operation")
    package = relationship("Package")
    status = relationship("Status")


class PackageSource(Base):
    __tablename__ = "package_source"

    id = Column(
        BigInteger,
        primary_key=True,
        server_default=text("nextval('package_source_id_seq'::regclass)"),
    )
    source_id = Column(ForeignKey("source.id"))
    package_id = Column(ForeignKey("package.id"))
    date_created = Column(DateTime(True), nullable=False, server_default=text("now()"))
    date_modified = Column(DateTime(True), nullable=False, server_default=text("now()"))

    package = relationship("Package")
    source = relationship("Source")


class WsusItem(Base):
    __tablename__ = "wsus_item"

    id = Column(
        BigInteger,
        primary_key=True,
        server_default=text("nextval('wsus_item_id_seq'::regclass)"),
    )
    source_id = Column(ForeignKey("source.id"))
    sha1 = Column(CHAR(40), nullable=False)
    archive_path = Column(String(256))
    update_type = Column(String(250))
    local_update_id = Column(Integer)
    local_update_guid = Column(UUID)
    title = Column(String)
    filename = Column(String)
    download_url = Column(String)
    product_family = Column(String(256))
    update_classification = Column(String(256))
    kb_article_id = Column(String(16))
    security_bulletin_id = Column(String(16))
    more_info_url = Column(String)
    date_created = Column(DateTime(True), nullable=False, server_default=text("now()"))
    date_modified = Column(DateTime(True), nullable=False, server_default=text("now()"))

    source = relationship("Source")
