import enum

from sqlalchemy import func
from sqlalchemy.orm import Load
from sqlalchemy.sql import text

from tic.shared.ubs.enums.cb_partners import CbPartner
from tic.shared.ubs.models.models import (
    FileIndex,
    FileDetail,
    Magic,
    FileDetailsMagicLink,
    FileDetailsCustLink,
    Cust,
    CustIxLink,
    FileDetailsPartnerLink,
    Metadatum,
    CustFilepath,
    Filepath,
    CustDeviceLink,
    CustDeviceLink1,
    CustDeviceLink3,
    Device,
    SignatureResultLink,
)

from tic.shared.ubs.util import hash_to_binary, binary_to_hash
from tic.shared.utils.date_util import utc_now
from tic.shared.utils.hash_util import calculate_sha256
from tic.shared.ubs.dtos.customer_access import CustLinkDetails
from tic.shared.ubs.dtos.file_details_and_magic import FileDetailsMagicDto
from tic.shared.ubs.repository.ubs_db_base import UbsDbRepositoryBase


class TableCustDeviceLink(enum.Enum):
    cust_device_link_1 = "cust_device_link_1"
    cust_device_link = "cust_device_link"
    cust_device_link_3 = "cust_device_link_3"

    def __str__(self):
        return str(self.value)


class UbsDbRepository(UbsDbRepositoryBase):
    def __init__(
        self,
        logger,
        writer_conn_string,
        reader_conn_string,
        connection_args=None,
        statement_timeout=1000,
    ):
        super(UbsDbRepository, self).__init__(
            logger,
            writer_conn_string,
            reader_conn_string,
            connection_args,
            statement_timeout,
        )

    # TODO: make sure readers are doing reads, writers are doing writes
    def get_file_details_and_magic(self, sha256, reader):
        """
        Gets the details about what's in the binary store for a given sha256
        @param sha256: The sha256 of the execution, for which you want to look up the stored binary details
        @param reader: Reader session param, ubs_reader_session instance
        @return: The size of the file in the store, whether the file was truncated, and the magic of the stored file.
        """
        result = (
            reader.query(FileIndex, FileDetail, Magic)
            .join(FileDetail, FileIndex.id == FileDetail.file_index_id)
            .join(
                FileDetailsMagicLink,
                FileDetail.id == FileDetailsMagicLink.file_details_id,
            )
            .join(Magic, FileDetailsMagicLink.magic_id == Magic.id)
            .options(
                Load(FileDetail).load_only(
                    FileDetail.size, FileDetail.truncated, FileDetail.test
                ),
                Load(Magic).load_only(Magic.value),
            )
            .filter(FileIndex.sha256 == hash_to_binary(sha256))
            .first()
        )

        if result is None:
            return None, None, None, None

        return (
            result.FileDetail.size,
            result.FileDetail.truncated,
            result.FileDetail.test,
            result.Magic.value,
        )

    def get_upload_history_for_file_and_partner(self, sha256, partner, reader):
        """
        Gets the upload history of a file to a given partner.
        @param sha256: the sha256 to check
        @param partner: A CbPartner object.
        @param reader: Reader session param, ubs_reader_session instance
        @return: The link if one exists, else None
        """

        return (
            reader.query(FileDetailsPartnerLink)
            .filter_by(partner_id=partner.value)
            .join(FileDetailsPartnerLink.file_details)
            .join(FileDetail.file_index)
            .filter_by(sha256=hash_to_binary(sha256))
            .first()
        )

    def put_partner_upload(self, sha256, partner, partner_status, writer):
        """
        Puts a record of a partner upload attempt
        @param sha256: the file we've shared
        @param partner: the partner we've shared with (CbPartner expected)
        @param partner_status: the status the partner gave us (enum expected)
        @param writer: Writer session param, ubs_writer_session instance
        @return:
        """

        existing_link = self.get_upload_history_for_file_and_partner(
            sha256, partner, writer
        )
        if existing_link is not None:
            existing_link.response_state = partner_status.value
            existing_link.date_modified = utc_now()
            writer.add(existing_link)
        else:
            writer.add(
                FileDetailsPartnerLink(
                    partner_id=partner.value,
                    response_state=partner_status.value,
                    file_details=writer.query(FileDetail)
                    .join(FileIndex, FileDetail.file_index_id == FileIndex.id)
                    .filter_by(sha256=hash_to_binary(sha256))
                    .first(),
                )
            )

    def put_upload_metadata(self, cust, file_index, file_details, magic_str, writer):
        """
        Inserts all records for an uploaded object
        @param cust: The customer record
        @param file_index: The file index object
        @param file_details: The file details of the upload
        @param magic_str: The magic string of the uploaded file
        @param writer: Writer session param, ubs_writer_session instance
        @return:
        """
        file_index = self.insert_file_index(writer, file_index)
        file_details = self.insert_update_file_details(writer, file_index, file_details)
        self.insert_update_magic(writer, file_details, magic_str)
        self.insert_update_cust_and_link(
            writer, file_index, cust, file_details=file_details
        )

    @staticmethod
    def get_file_index(session, sha256):
        """
        Gets existing file index if it exists
        @param session: sqlalchemy session to query on
        @param sha256: sha256 of the record
        @return: the file_index object from the databae if it exists else None
        """
        return session.query(FileIndex).filter_by(sha256=sha256).first()

    def insert_file_index(self, session, file_index):
        """
        Inserts a file_index into the UBS db if one does not exist.
        @param session: sqlalchemy session to query on
        @param file_index: a FileIndex object
        @return: the id of the file_index entry in the db
        """
        db_file_index = self.get_file_index(session, file_index.sha256)

        # if no entry where sha256
        if db_file_index is None:
            # update md5_id in case of md5 collision
            file_index.md5_id = (
                session.query(FileIndex).filter_by(md5=file_index.md5).count()
            )
            session.add(file_index)
            db_file_index = (
                session.query(FileIndex).filter_by(sha256=file_index.sha256).first()
            )

        return db_file_index

    @staticmethod
    def insert_update_file_details(session, file_index, file_details):
        """
        Inserts a file_index into the UBS db if one does not exist.
        @param session: sqlalchemy session to run queries on
        @param file_index: the file_index object on the session that this should link to
        @param file_details: the FileDetails object to insert
        @return: the id of the file_details entry in the db
        """
        existing_object = (
            session.query(FileDetail).filter_by(file_index=file_index).first()
        )
        if (
            existing_object is not None
        ):  # we only get here if we've already determined that we want the file
            existing_object.size = file_details.size
            existing_object.truncated = file_details.truncated
            session.add(existing_object)
            return existing_object
        else:
            file_details.file_index = file_index
            session.add(file_details)
            return session.query(FileDetail).filter_by(file_index=file_index).first()

    @staticmethod
    def insert_update_filepath_testing_purposes(session, filepath):
        """
        Method used for testing that inserts a filepath into the UBS db if one does not exist.
        @param session: sqlalchemy session to run queries on
        @param filepath: the filepath to be inserted
        @return: the filepath entry in the db
        """
        existing_object = (
            session.query(Filepath).filter_by(sha256_path=filepath.sha256_path).first()
        )
        if existing_object is None:
            session.add(filepath)
            existing_object = (
                session.query(Filepath)
                .filter_by(sha256_path=filepath.sha256_path)
                .first()
            )

        return existing_object

    @staticmethod
    def insert_update_device_testing_purposes(session, device):
        """
        Method used for testing that inserts a device into the UBS db if one does not exist.
        @param session: sqlalchemy session to run queries on
        @param device: the device to be inserted
        @return: the device entry in the db
        """
        existing_object = (
            session.query(Device).filter_by(device_name=device.device_name).first()
        )
        if existing_object is None:
            session.add(device)
            existing_object = (
                session.query(Device).filter_by(device_name=device.device_name).first()
            )

        return existing_object

    @staticmethod
    def insert_update_magic(session, file_details, magic_str):
        """
        Inserts magic and links to file_details
        @param session: the session to query on
        @param file_details: the file details object on the session to link to
        @param magic_str: the magic string to insert if necessary
        """
        magic_truncated = magic_str[:1024]
        magic_sha256 = hash_to_binary(calculate_sha256(magic_truncated))
        existing_magic = (
            session.query(Magic).filter_by(sha256_value=magic_sha256).first()
        )

        # handle adding magic if necessary
        if existing_magic is None:
            magic_obj = Magic(value=magic_truncated, sha256_value=magic_sha256)
            session.add(magic_obj)
            existing_magic = (
                session.query(Magic).filter_by(sha256_value=magic_sha256).first()
            )
            existing_link = None
        else:
            existing_link = (
                session.query(FileDetailsMagicLink)
                .filter_by(file_details=file_details)
                .first()
            )

        # handle magic/file_details link if necessary now that we have a magic object on the session
        if existing_link is not None:
            existing_link.magic_id = existing_magic.id
            session.add(existing_link)
        else:
            session.add(
                FileDetailsMagicLink(magic=existing_magic, file_details=file_details)
            )

    def insert_update_cust_and_link(
        self, session, file_index, cust, file_details=None, date_override=None
    ):
        """
        Inserts into cust if necessary and produces cust_ix_link
        @param session: the session to query on
        @param file_index: the file index on the session to link to
        @param cust: the cust object to insert or link
        @param file_details: optional file_details to link cust to
        @param date_override: option date to override the normal utc.now() timestamp
        """

        existing_cust = (
            session.query(Cust).filter_by(search_key=cust.search_key).first()
        )

        if existing_cust is None:
            # add customer if necessary
            session.add(cust)
            existing_cust = (
                session.query(Cust).filter_by(search_key=cust.search_key).first()
            )
            existing_link = None
        else:
            # if customer exists check for existing link
            existing_link = (
                session.query(CustIxLink)
                .filter_by(cust=existing_cust)
                .filter_by(file_index=file_index)
                .first()
            )

        if existing_link is None:
            # if no link, add a link.
            if date_override:
                session.add(
                    CustIxLink(
                        cust=existing_cust,
                        file_index=file_index,
                        date_created=date_override,
                    )
                )
            else:
                session.add(CustIxLink(cust=existing_cust, file_index=file_index))

        if file_details:
            self._insert_cust_file_details_link(session, file_details, existing_cust)

    @staticmethod
    def _insert_cust_file_details_link(session, file_details, cust):
        """

        @param session: the session to query on
        @param file_details: the file that was uploaded
        @param cust: the original uploader to reference
        @return:
        """

        # always add a new row (date_created will vary)
        session.add(FileDetailsCustLink(cust=cust, file_details=file_details))

    def get_file_details(self, sha256, reader):
        """
        Returns the File Details object corresponding to the sha256, otherwise returns None
        @param sha256: The file hash
        @param reader: Reader session param, ubs_reader_session instance
        @return: The corresponding File Details objects, otherwise None
        """

        file_details = (
            reader.query(FileDetail)
            .join(FileIndex, FileDetail.file_index_id == FileIndex.id)
            .filter(FileIndex.sha256 == hash_to_binary(sha256))
            .first()
        )

        return file_details

    def get_file_details_bulk(self, sha256s, reader):
        """
        Returns the File Details objects corresponding to the sha256s, otherwise returns None
        @param sha256s: The file hashes
        @param reader: Reader session param, ubs_reader_session instance
        @return: A dict of sha256s to File Details objects, or None if no details exist
        """

        bin_sha256s = {hash_to_binary(sha256) for sha256 in sha256s}

        file_details = (
            reader.query(FileDetail)
            .join(FileIndex, FileDetail.file_index_id == FileIndex.id)
            .filter(FileIndex.sha256.in_(bin_sha256s))
            .all()
        )

        results = {}

        for file_detail in file_details:
            results[binary_to_hash(file_detail.file_index.sha256).lower()] = file_detail

        return results

    def get_file_details_and_magic_bulk(self, sha256s, reader):
        """
        Returns the File Details objects corresponding to the sha256s, otherwise returns None
        @param sha256s: The file hashes
        @param reader: Reader session param, ubs_reader_session instance
        @return: A dict of sha256s to File Details objects, or None if no details exist
        """

        bin_sha256s = {hash_to_binary(sha256) for sha256 in sha256s}

        ret = (
            reader.query(FileDetail, FileDetailsMagicLink)
            .join(FileIndex, FileIndex.id == FileDetail.file_index_id)
            .filter(FileIndex.sha256.in_(bin_sha256s))
            .filter(FileDetail.id == FileDetailsMagicLink.file_details_id)
            .all()
        )

        results = {}

        for obj in ret:
            results[
                binary_to_hash(obj[0].file_index.sha256).lower()
            ] = FileDetailsMagicDto(file_details=obj[0], file_details_magic_link=obj[1])
        return results

    def get_file_details_magic(self, sha256, reader):
        """
        Returns the File Details and Magic corresponding to the sha256, otherwise returns None
        @param sha256: The file hash
        @param reader: Reader session param, ubs_reader_session instance
        @return: A dict of sha256s to File Details objects, or None if no details exist
        """

        result = (
            reader.query(FileIndex, FileDetail, Magic)
            .join(FileDetail, FileIndex.id == FileDetail.file_index_id)
            .join(
                FileDetailsMagicLink,
                FileDetail.id == FileDetailsMagicLink.file_details_id,
            )
            .join(Magic, FileDetailsMagicLink.magic_id == Magic.id)
            .options(
                Load(FileDetail).load_only(FileDetail.sha256, FileDetail.s3_key),
                Load(Magic).load_only(Magic.value),
            )
            .filter(FileIndex.sha256 == hash_to_binary(sha256))
            .first()
        )

        if not result:
            return {}

        return {
            "sha256_truncated": binary_to_hash(result.FileDetail.sha256),
            "s3_key": result.FileDetail.s3_key,
            "magic": result.Magic.value,
        }

    def get_metadata(self, sha256, reader):
        """
        Returns the Metadatum object corresponding to the sha256, otherwise returns None
        @param sha256: The file hash
        @param reader: Reader session param, ubs_reader_session instance
        @return: The corresponding Metadatum object, otherwise None
        """

        metadata = (
            reader.query(Metadatum)
            .join(FileIndex, Metadatum.file_index_id == FileIndex.id)
            .filter(FileIndex.sha256 == hash_to_binary(sha256))
            .first()
        )

        return metadata

    def get_metadata_bulk(self, sha256s, reader):
        """
        Returns the Metadatum objects corresponding to the sha256s, otherwise returns None
        @param sha256s: The file hashes
        @param reader: Reader session param, ubs_reader_session instance
        @return: A dict of sha256s to Metadatum objects
        """

        bin_sha256s = {hash_to_binary(sha256) for sha256 in sha256s}

        metadata = (
            reader.query(Metadatum, FileIndex.sha256)
            .join(FileIndex, Metadatum.file_index_id == FileIndex.id)
            .filter(FileIndex.sha256.in_(bin_sha256s))
            .options(
                Load(Metadatum).load_only(
                    Metadatum.charset_id,
                    Metadatum.comments,
                    Metadatum.file_description,
                    Metadatum.file_version,
                    Metadatum.lang_id,
                    Metadatum.original_filename,
                    Metadatum.private_build,
                    Metadatum.product_description,
                    Metadatum.product_version,
                    Metadatum.special_build,
                )
            )
        )

        results = []

        for metadatum in metadata:
            metadatum[0].sha256 = binary_to_hash(metadatum[1])
            results.append(metadatum[0])

        return results

    def get_file_access_bulk(self, org_search_key, sha256s, reader):
        """
        Returns the sha256s which the customer has access to, otherwise returns []
        @param org_search_key: Search key for the customer in question
        @param sha256s: The file hashes
        @param reader: Reader session param, ubs_reader_session instance
        @return: A list of sha256s that the org has access to
        """

        bin_sha256s = {hash_to_binary(sha256) for sha256 in sha256s}

        found_sha256s = (
            reader.query(FileIndex.sha256)
            .join(CustIxLink, FileIndex.id == CustIxLink.file_index_id)
            .join(Cust, Cust.id == CustIxLink.cust_id)
            .filter(Cust.search_key == org_search_key)
            .filter(FileIndex.sha256.in_(bin_sha256s))
            .all()
        )

        return [binary_to_hash(record[0]) for record in found_sha256s]

    def get_metadata_version_bulk(self, sha256s, reader):
        """
        Returns sha256s of binaries with out of date metadata
        @param sha256s: The file hashes
        @param reader: Reader session param, ubs_reader_session instance
        @return: sha256s that are out of date
        """
        bin_sha256s = {hash_to_binary(sha256) for sha256 in sha256s}
        metadata = (
            reader.query(Metadatum, FileIndex.sha256)
            .with_entities(
                FileIndex.sha256, Metadatum.internal_version, Metadatum.policy_version
            )
            .join(Metadatum.file_index)
            .filter(FileIndex.sha256.in_(bin_sha256s))
        )

        results = dict()

        for metadatum in metadata:
            results[binary_to_hash(metadatum.sha256)] = {
                "internal_version": metadatum.internal_version,
                "policy_version": metadatum.policy_version,
            }

        return results

    def get_total_file_path_count(self, cust_ix_link, reader):
        """
        Returns the total number of file paths that have been observed, by this customer, for this file.
        @param cust_ix_link: The Customer_ix_link of the customer
        @param reader: Reader session param, ubs_reader_session instance
        @return: The total number of file paths that have been observed, by this customer, for this file.
        """

        count = (
            reader.query(func.sum(CustFilepath.num_occurrences))
            .join(CustIxLink, CustFilepath.cust_ix_link_id == CustIxLink.id)
            .filter(CustFilepath.cust_ix_link == cust_ix_link)
            .first()
        )

        if count[0] is None:
            return 0
        return count[0]

    def get_file_path_count(self, cust_ix_link, reader):
        """
        Returns the total number of unique file paths that have been observed, by this customer, for this file.
        @param cust_ix_link: The Customer_ix_link of the customer
        @param reader: Reader session param, ubs_reader_session instance
        @return: The total number of unique file paths that have been observed, by this customer, for this file.
        """

        count = (
            reader.query(CustFilepath)
            .join(CustIxLink, CustFilepath.cust_ix_link_id == CustIxLink.id)
            .filter(CustFilepath.cust_ix_link == cust_ix_link)
            .count()
        )

        return count

    def get_file_path_array(self, cust_ix_link, num_results, reader):
        """
        Returns an array of cust filepaths. The length of the array is limited based off the provided input parameter.
        @param cust_ix_link: The Customer_ix_link of the customer
        @param num_results: The length of the array to return
        @param reader: Reader session param, ubs_reader_session instance
        @return: The cust file path array. The length of the array is limited based off the provided input parameter.
        """

        file_paths = (
            (
                reader.query(CustFilepath)
                .join(CustIxLink, CustFilepath.cust_ix_link_id == CustIxLink.id)
                .filter(CustFilepath.cust_ix_link == cust_ix_link)
                .order_by(CustFilepath.num_occurrences.desc())
            )
            .limit(num_results)
            .all()
        )

        if file_paths is None:
            file_paths = []

        return file_paths

    def get_num_devices(
        self, cust_ix_link, reader, cdl=TableCustDeviceLink.cust_device_link
    ):
        """
        Returns the total number of devices, for this customer, that have observed this file.
        @param cust_ix_link: The Customer_ix_link of the customer
        @param reader: Reader session param, ubs_reader_session instance
        @return: The total number of devices, for this customer, that have observed this file.
        """
        count = {
            TableCustDeviceLink.cust_device_link: (
                reader.query(CustDeviceLink)
                .join(CustIxLink, CustDeviceLink.cust_ix_link_id == CustIxLink.id)
                .filter(CustDeviceLink.cust_ix_link == cust_ix_link)
                .count()
            ),
            TableCustDeviceLink.cust_device_link_1: (
                reader.query(CustDeviceLink1)
                .join(CustIxLink, CustDeviceLink1.cust_ix_link_id == CustIxLink.id)
                .filter(CustDeviceLink1.cust_ix_link == cust_ix_link)
                .count()
            ),
            TableCustDeviceLink.cust_device_link_3: (
                reader.query(CustDeviceLink3)
                .join(CustIxLink, CustDeviceLink3.cust_ix_link_id == CustIxLink.id)
                .filter(CustDeviceLink3.cust_ix_link == cust_ix_link)
                .count()
            ),
        }[cdl]

        return count

    def get_first_seen_device(
        self, cust_ix_link, reader, cdl=TableCustDeviceLink.cust_device_link
    ):
        """
        Returns the first seen device for this customer, that have observed this file.
        @param cust_ix_link: The Customer_ix_link of the customer
        @param reader: Reader session param, ubs_reader_session instance
        @return: The Device that first saw this file.
        """

        first_device = {
            TableCustDeviceLink.cust_device_link: (
                reader.query(CustDeviceLink)
                .join(CustIxLink, CustDeviceLink.cust_ix_link_id == CustIxLink.id)
                .filter(CustDeviceLink.cust_ix_link == cust_ix_link)
                .order_by(CustDeviceLink.first_seen.asc())
                .first()
            ),
            TableCustDeviceLink.cust_device_link_1: (
                reader.query(CustDeviceLink1)
                .join(CustIxLink, CustDeviceLink1.cust_ix_link_id == CustIxLink.id)
                .filter(CustDeviceLink1.cust_ix_link == cust_ix_link)
                .order_by(CustDeviceLink1.first_seen.asc())
                .first()
            ),
            TableCustDeviceLink.cust_device_link_3: (
                reader.query(CustDeviceLink3)
                .join(CustIxLink, CustDeviceLink3.cust_ix_link_id == CustIxLink.id)
                .filter(CustDeviceLink3.cust_ix_link == cust_ix_link)
                .order_by(CustDeviceLink3.first_seen.asc())
                .first()
            ),
        }[cdl]

        return first_device

    # Note - this function is getting the last first seen device
    def get_last_seen_device(
        self, cust_ix_link, reader, cdl=TableCustDeviceLink.cust_device_link
    ):
        """
        Returns the last first seen device for this customer, that have observed this file.
        @param cust_ix_link: The Customer_ix_link of the customer
        @param reader: Reader session param, ubs_reader_session instance
        @return: The Device that last saw this file. Note: actually last first seen
        """

        last_first_device = {
            TableCustDeviceLink.cust_device_link: (
                reader.query(CustDeviceLink)
                .join(CustIxLink, CustDeviceLink.cust_ix_link_id == CustIxLink.id)
                .filter(CustDeviceLink.cust_ix_link == cust_ix_link)
                .order_by(CustDeviceLink.first_seen.desc())
                .first()
            ),
            TableCustDeviceLink.cust_device_link_1: (
                reader.query(CustDeviceLink1)
                .join(CustIxLink, CustDeviceLink1.cust_ix_link_id == CustIxLink.id)
                .filter(CustDeviceLink1.cust_ix_link == cust_ix_link)
                .order_by(CustDeviceLink1.first_seen.desc())
                .first()
            ),
            TableCustDeviceLink.cust_device_link_3: (
                reader.query(CustDeviceLink3)
                .join(CustIxLink, CustDeviceLink3.cust_ix_link_id == CustIxLink.id)
                .filter(CustDeviceLink3.cust_ix_link == cust_ix_link)
                .order_by(CustDeviceLink3.first_seen.desc())
                .first()
            ),
        }[cdl]

        return last_first_device

    def get_total_signature_count(self, cust_ix_link, reader):
        """
        Returns the total number of signatures that have been observed, by this customer, for this file.
        @param cust_ix_link: The Customer_ix_link of the customer
        @param reader: Reader session param, ubs_reader_session instance
        @return: The total number of signatures that have been observed, by this customer, for this file.
        """

        count = (
            reader.query(func.sum(SignatureResultLink.num_occurrences))
            .join(CustIxLink, SignatureResultLink.cust_ix_link_id == CustIxLink.id)
            .filter(SignatureResultLink.cust_ix_link == cust_ix_link)
            .first()
        )

        if count[0] is None:
            return 0
        return count[0]

    def get_signature_count(self, cust_ix_link, reader):
        """
        Returns the total number of unique signatures that have been observed, by this customer, for this file.
        @param cust_ix_link: The Customer_ix_link of the customer
        @param reader: Reader session param, ubs_reader_session instance
        @return: The total number of unique signatures that have been observed, by this customer, for this file.
        """

        count = (
            reader.query(SignatureResultLink)
            .join(CustIxLink, SignatureResultLink.cust_ix_link_id == CustIxLink.id)
            .filter(SignatureResultLink.cust_ix_link == cust_ix_link)
            .count()
        )

        return count

    def get_signature_array(self, cust_ix_link, num_results, reader):
        """
        Returns an array of signature_result_link object. The length of the
        array is limited based off the provided num_results input parameter.
        @param cust_ix_link: The Customer_ix_link of the customer
        @param num_results: The length of the array to return
        @param reader: Reader session param, ubs_reader_session instance
        @return: Returns an array of signature_result_link object of length num_results
        """

        signatures = (
            (
                reader.query(SignatureResultLink)
                .join(CustIxLink, SignatureResultLink.cust_ix_link_id == CustIxLink.id)
                .filter(SignatureResultLink.cust_ix_link == cust_ix_link)
                .order_by(SignatureResultLink.num_occurrences.desc())
            )
            .limit(num_results)
            .all()
        )

        if signatures is None:
            signatures = []

        return signatures

    @staticmethod
    def get_cust_device_link_tables(reader):
        """
        @param reader: reader session
        @return: current and next cust_device_link table names from ubs_table_rotation
        """
        ret = (
            reader.execute(
                text(
                    "select * from ubs_table_rotation where table_name='cust_device_link'"
                )
            )
            .mappings()
            .all()
        )

        return ret[0]["current"], ret[0]["next"]

    @staticmethod
    def get_duplicate_device_links(
        reader, tuple_mapping, cdl=TableCustDeviceLink.cust_device_link
    ):
        """
        @param reader: reader session parameter
        @param tuple_mapping: DeviceInsert objects to be queried and checked
        @param cdl: cust_device_link table that the dedupe function should check
        @return: a tuple of (hash, org_key, device_id) of cust device links that already exist
        """

        statement = {
            TableCustDeviceLink.cust_device_link: "SELECT * FROM select_duplicate_cust_device_links_v3((:item)::hash_cust_device_v3[])",
            TableCustDeviceLink.cust_device_link_1: "SELECT * FROM select_duplicate_cust_device_links_1_v3((:item)::hash_cust_device_v3[])",
            TableCustDeviceLink.cust_device_link_3: "SELECT * FROM select_duplicate_cust_device_links_3_v3((:item)::hash_cust_device_v3[])",
        }[cdl]

        ret = reader.execute(
            text(statement),
            {
                "item": [
                    (
                        sha256,
                        org_key,
                        device_id,
                        device_name,
                    )
                    for sha256, org_key, device_id, device_name in tuple_mapping
                ]
            },
        ).fetchall()

        return {tuple(row) for row in ret}

    @staticmethod
    def insert_file_index_cust_partner_link(
        session, sha256s, org_key, partner: CbPartner
    ):
        """
        @param session: the session to query on
        @param sha256s: the hashes of the files that we want to link on
        @param org_key: the org_key of the cust that requested the file
        @param partner: the partner who the org key is requesting data from - CbPartner enum
        @return:
        """

        session.execute(
            text(
                "SELECT insert_file_index_cust_partner_link("
                "(:item)::file_index_cust_partner[]) LIMIT 1"
            ),
            {
                "item": [
                    (hash_to_binary(sha256), org_key, partner.value)
                    for sha256 in sha256s
                ]
            },
        )

    @staticmethod
    def add_file_indexes(session, index_customer_metadata):
        """
        Executes the index_customer_metadata stored procedure.

        :param session: SqlAlchemy session
        :param index_customer_metadata: List of IndexCustomerMetadata DTOs.
        """

        session.execute(
            text(
                "SELECT insert_file_index_and_cust_link("
                "(:item)::index_customer_metadata[]) LIMIT 1"
            ),
            {"item": [x.to_tuple() for x in index_customer_metadata]},
        )
        session.commit()

    @staticmethod
    def add_signatures(session, signature_data):
        """
        Executes the insert_signature stored procedure.

        :param session: SqlAlchemy session
        :param signature_data: List of SignatureData DTOs.
        """

        session.execute(
            text(
                "SELECT insert_signature_v4(" "(:item)::signature_insert_v4[]) LIMIT 1"
            ),
            {"item": [x.to_tuple() for x in signature_data]},
        )
        session.commit()

    @staticmethod
    def add_static_metadata(static_metadata_insert, writer):
        """
        Executes the insert_static_metadata stored procedure.

        :param static_metadata_insert: List of StaticMetadataInsert objects.
        :param session: optional SqlAlchemy session
        """
        writer.execute(
            text(
                "SELECT insert_static_metadata_v2((:item)::static_metadata_insert_v3[]) LIMIT 1"
            ),
            {"item": [x.to_tuple() for x in static_metadata_insert]},
        )
        writer.commit()

    @staticmethod
    def add_devices(device_insert, writer, cdl=TableCustDeviceLink.cust_device_link):
        """
        Executes the insert_static_metadata stored procedure.

        :param device_insert: List of DeviceInsert objects.
        :param session: optional SqlAlchemy session
        """
        statement = {
            TableCustDeviceLink.cust_device_link: "SELECT insert_device_v3((:item)::device_insert_v3[]) LIMIT 1",
            TableCustDeviceLink.cust_device_link_1: "SELECT insert_device_1_v3((:item)::device_insert_v3[]) LIMIT 1",
            TableCustDeviceLink.cust_device_link_3: "SELECT insert_device_3_v3((:item)::device_insert_v3[]) LIMIT 1",
        }[cdl]

        ret = writer.execute(
            text(statement),
            {"item": [x.to_tuple() for x in device_insert]},
        )
        writer.commit()

        cleaned_ret = list(ret)
        hashes = list()
        for row in cleaned_ret:
            hashes.extend([binary_to_hash(h) for h in row[0]])

        return hashes

    @staticmethod
    def add_file_path(session, filepath_insert):
        """
        Executes the index_customer_metadata stored procedure.

        :param session: SqlAlchemy session
        :param filepath_insert: List of FilepathInsert DTOs.
        """

        session.execute(
            text("SELECT insert_filepath_v2(" "(:item)::filepath_insert_v2[]) LIMIT 1"),
            {"item": [x.to_tuple() for x in filepath_insert]},
        )
        session.commit()

    @staticmethod
    def get_batch_cust_access(search_key, sha256_strs, reader):
        """
        batch call to return links that exist and that the customer has access to
        @param search_key: Search key for the customer in question
        @param sha256_strs: String of the hashes that the customer wants to access
        @param reader: reader session parameter
        @return: A dictionary of sha256s as a key, with their respective values being a dictionary
                 with their corresponding cust id, link id, and file index id
        """

        bin_sha256s = {hash_to_binary(sha256_str) for sha256_str in sha256_strs}

        links = (
            reader.query(
                CustIxLink.cust_id,
                CustIxLink.id,
                CustIxLink.file_index_id,
                FileIndex.sha256,
            )
            .join(Cust, Cust.id == CustIxLink.cust_id)
            .outerjoin(FileIndex, FileIndex.id == CustIxLink.file_index_id)
            .filter(Cust.search_key == search_key)
            .filter(FileIndex.sha256.in_(bin_sha256s))
            .all()
        )
        results = []

        for link in links:
            results.append(
                CustLinkDetails(
                    sha256=binary_to_hash(link[3]),
                    cust_id=link[0],
                    cust_ix_link_id=link[1],
                    file_index_id=link[2],
                )
            )
        return results
