from io import BytesIO
import os
from shutil import copyfileobj
from zipfile import ZipFile

from tic.shared.aws.kms_client import KmsClient
from tic.shared.aws.s3_client import S3Client


class ResponseFileStoreRepository(object):
    def __init__(self, boto_session):
        self.s3_client = S3Client(boto_session)
        self.kms_client = KmsClient(boto_session)

    def download_and_extract_in_memory(
        self, s3_bucket_name, s3_key_name, encryption_key_alias=None
    ):
        # Encrypted
        if encryption_key_alias:
            s3_contents = self.kms_client.decrypt_data_with_boto(
                s3_bucket_name, s3_key_name
            )

        # Unencrypted
        else:
            with BytesIO() as b:
                self.s3_client.download_fileobj(
                    file_obj=b, bucket_name=s3_bucket_name, key_name=s3_key_name
                )
                s3_contents = b.getvalue()

        with BytesIO(s3_contents) as contents_io:
            zf = ZipFile(contents_io)
            extracted_files = [
                {"contents": zf.open(filename).read(), "name": filename}
                for filename in zf.namelist()
            ]

        return extracted_files

    @staticmethod
    def extract_sample_file_in_memory(data_in_memory):
        # CB sends us filedata and metadata
        # B9P sends us just one file in the zip file embedded in a deep thing
        for f in data_in_memory:
            if f["name"].endswith("filedata"):
                return f["contents"]
        return data_in_memory[0]["contents"]

    def get_sample_file_in_memory(
        self, s3_bucket_name, s3_key_name, encryption_key_alias=None
    ):
        return self.extract_sample_file_in_memory(
            self.download_and_extract_in_memory(
                s3_bucket_name, s3_key_name, encryption_key_alias
            )
        )

    def download_to_disk(
        self,
        download_s3_bucket,
        download_s3_file_path,
        temp_folder,
        encryption_key_alias=None,
    ):
        # download file
        if not os.path.exists(temp_folder):
            os.makedirs(temp_folder)

        folder_name = temp_folder
        zip_file_name = folder_name + "/temp.zip"

        if os.path.exists(zip_file_name):
            os.remove(zip_file_name)
        if encryption_key_alias:
            self.kms_client.decrypt_file_with_boto(
                zip_file_name, download_s3_bucket, download_s3_file_path
            )

        else:
            self.s3_client.download_file(
                zip_file_name, download_s3_bucket, download_s3_file_path
            )

        return zip_file_name

    @staticmethod
    def extract_zip_on_disk(temp_folder, zip_file_name):
        # extract the zip
        if not os.path.exists(temp_folder):
            os.makedirs(temp_folder)

        extracted_files = []
        with ZipFile(zip_file_name) as zf:
            for file_in_zip in zf.namelist():
                if not (file_name := os.path.basename(file_in_zip)):
                    continue

                source = zf.open(file_in_zip)
                target = open(os.path.join(temp_folder, file_name), "wb")
                with source, target:
                    copyfileobj(source, target)
                extracted_files.append(os.path.join(temp_folder, file_name))

        return extracted_files

    @staticmethod
    def find_sample_file_on_disk(extracted_files):
        # CB sends us filedata and metadata
        # B9P sends us just one file in the zip file embedded in a deep thing
        for f in extracted_files:
            if f.endswith("filedata"):
                return f
        return extracted_files[0]
